/**
 * {}
 * @author yutent<yutent.io@gmail.com>
 * @date 2024/03/26 10:37:00
 */
import {} from './lib/constants.js'
import { $, uuid, h, preload, hyphen } from './utils.js'

export class Component {
  #node
  type
  id
  defs

  constructor(el) {
    this.#node = el

    let _type = el.tagName.toLowerCase()

    if (_type !== 'defs') {
      this.id = uuid(_type)
      el.id = this.id
    }

    this.type = _type

    if (_type === 'svg') {
      let defs = this.select('defs')
      if (!defs) {
        defs = this.#create('defs')
      }
      this.defs = defs
    }
  }

  get node() {
    return this.#node
  }

  get textContent() {
    return this.#node.textContent
  }

  set textContent(val) {
    this.#node.textContent = val
  }

  get children() {
    return Array.from(this.#node.children)
  }

  #create(name, attr, child) {
    let node = h(name, attr, child)
    this.#node.appendChild(node)
    return new Component(node)
  }

  append(...elems) {
    for (let el of elems) {
      this.#node.appendChild(el instanceof Component ? el.node : el)
    }
    return this
  }

  select(selector) {
    let node = $(selector, this.#node)
    if (node) {
      return new Component(node)
    }
    return null
  }

  attr(key, val) {
    let node = this.node
    let out = {}

    if (key) {
      let attrs = key
      if (typeof key === 'string') {
        if (val === void 0) {
          return node.getAttribute(key)
        } else {
          attrs = { [key]: val }
        }
      }

      h(node, attrs)
      return this
    }

    // 无任何参数时, 返回节点所有的属性
    for (let it of Array.from(node.attributes)) {
      out[it.nodeName] = it.nodeValue
    }

    return out
  }

  move(x, y) {
    this.attr({ x, y })
    return this
  }

  size(width, height) {
    this.attr({ width, height })
    return this
  }

  transform(transform) {
    this.attr({ transform })
    return this
  }

  /* ------------------------------ */

  /**
   * 引入defs中的组件
   */
  use(id) {
    if (id instanceof Component) {
      id = id.id
    } else if (id.startsWith('#')) {
      id = id.slice(1)
    }
    return this.#create('use', { 'xlink:href': '#' + id })
  }

  g(...args) {
    let el = this.#create('g')
    if (args.length) {
      el.append(...args)
    }
    return el
  }

  style(content) {
    let node = this.select('style')
    if (node === null) {
      node = this.#create('style')
    }
    node.textContent += content
    return node
  }

  symbol(vx, vy, vw, vh) {
    return this.#create('symbol', { viewBox: [vx, vy, vw, vh].join(' ') })
  }

  rect(x, y, width, height) {
    return this.#create('rect', { x, y, width, height })
  }
  /**
   * 多边形, 传入多个坐标点
   */
  polygon(...points) {
    if (points.length) {
      points = points.flatMap(p => p.join(',')).join(' ')
      return this.#create('polygon', { points })
    }
    throw new Error('polygon() argument is undefined')
  }

  text(text = '', x = 0, y = 0) {
    return this.#create('text', { x, y }, text)
  }

  // 使用foreignObject实现文本自动换行等排版
  autoText(text = '', x = 0, y = 0, width, height) {
    return this.#create('foreignObject', { x, y, width, height }, text)
  }

  line(x1, y1, x2, y2, attr = { stroke: '#000', strokeWidth: 1 }) {
    return this.#create('line', { x1, y1, x2, y2, ...attr })
  }

  /**
   * 拆线, 传入多组坐标
   */
  polyline(...points) {
    if (points.length) {
      points = points.flatMap(p => p.join(',')).join(' ')
      return this.#create('polyline', { points })
    }
    throw new Error('polyline() argument is undefined')
  }

  circle(cx = 0, cy = 0, r = 1) {
    return this.#create('circle', { cx, cy, r })
  }

  /**
   * 椭圆
   * @param cx, cy  圆心坐标
   * @param rx, ry  X、Y轴半径
   */
  ellipse(cx, cy, rx, ry) {
    return this.#create('ellipse', { cx, cy, rx, ry })
  }

  image(href, { x, y, width, height } = {}) {
    let el = this.#create('image', {
      preserveAspectRatio: 'none',
      href,
      x,
      y,
      width,
      height
    })

    if (width === void 0 || height === void 0) {
      preload(href, function (img) {
        h(el.node, {
          width: img.naturalWidth,
          height: img.naturalHeight
        })
      })
    }
    return el
  }

  pattern(x, y, width, height, viewBox = []) {
    let props = { patternUnits: 'userSpaceOnUse', x, y, width, height }

    if (viewBox.length) {
      props.viewBox = viewBox.join(' ')
    } else {
      props.viewBox = [x, y, width, height].join(' ')
    }

    return this.#create('pattern', props)
  }

  path(d) {
    return this.#create('path', { d })
  }
}
